@@ -1,6 +1,6 @@
 ;;; exwm-floating.el --- Floating Module for EXWM  -*- lexical-binding: t -*-
 
-;; Copyright (C) 2015-2019 Free Software Foundation, Inc.
+;; Copyright (C) 2015-2024 Free Software Foundation, Inc.
 
 ;; Author: Chris Feng <chris.w.feng@gmail.com>
 
@@ -31,22 +31,18 @@
 
 (defgroup exwm-floating nil
   "Floating."
-  :version "25.3"
   :group 'exwm)
 
 (defcustom exwm-floating-setup-hook nil
-  "Normal hook run when an X window has been made floating, in the
-context of the corresponding buffer."
+  "Normal hook run when an X window has been made floating.
+This hook runs in the context of the corresponding buffer."
   :type 'hook)
 
 (defcustom exwm-floating-exit-hook nil
-  "Normal hook run when an X window has exited floating state, in the
-context of the corresponding buffer."
+  "Normal hook run when an X window has exited floating state.
+This hook runs in the context of the corresponding buffer."
   :type 'hook)
 
-(defvar exwm-floating--border-pixel nil
-  "Border pixel drawn around floating X windows.")
-
 (defcustom exwm-floating-border-color "navy"
   "Border color of floating windows."
   :type 'color
@@ -54,28 +50,28 @@ context of the corresponding buffer."
   :set (lambda (symbol value)
          (set-default symbol value)
          ;; Change border color for all floating X windows.
-         (exwm-floating--init-border)
-         (dolist (pair exwm--id-buffer-alist)
-           (with-current-buffer (cdr pair)
-             (when exwm--floating-frame
-               (xcb:+request exwm--connection
-                   (make-instance 'xcb:ChangeWindowAttributes
-                                  :window
-                                  (frame-parameter exwm--floating-frame
-                                                   'exwm-container)
-                                  :value-mask xcb:CW:BorderPixel
-                                  :border-pixel
-                                  exwm-floating--border-pixel)))))
          (when exwm--connection
-           (xcb:flush exwm--connection))))
+           (let ((border-pixel (exwm--color->pixel value)))
+             (when border-pixel
+               (dolist (pair exwm--id-buffer-alist)
+                 (with-current-buffer (cdr pair)
+                   (when exwm--floating-frame
+                     (xcb:+request exwm--connection
+                         (make-instance 'xcb:ChangeWindowAttributes
+                                        :window
+                                        (frame-parameter exwm--floating-frame
+                                                         'exwm-container)
+                                        :value-mask xcb:CW:BorderPixel
+                                        :border-pixel border-pixel)))))
+               (xcb:flush exwm--connection))))))
 
 (defcustom exwm-floating-border-width 1
   "Border width of floating windows."
-  :type '(integer
-          :validate (lambda (widget)
-                      (when (< (widget-value widget) 0)
-                        (widget-put widget :error "Border width is at least 0")
-                        widget)))
+  :type `(integer
+          :validate ,(lambda (widget)
+                       (when (< (widget-value widget) 0)
+                         (widget-put widget :error "Border width is at least 0")
+                         widget)))
   :initialize #'custom-initialize-default
   :set (lambda (symbol value)
          (let ((delta (- value exwm-floating-border-width))
@@ -104,11 +100,6 @@ context of the corresponding buffer."
            (when exwm--connection
              (xcb:flush exwm--connection)))))
 
-(defvar exwm-floating--border-colormap nil
-  "Colormap used by the border pixel.
-
-This is also used by X window containers.")
-
 ;; Cursors for moving/resizing a window
 (defvar exwm-floating--cursor-move nil)
 (defvar exwm-floating--cursor-top-left nil)
@@ -126,21 +117,22 @@ This is also used by X window containers.")
 (defvar exwm-workspace--current)
 (defvar exwm-workspace--frame-y-offset)
 (defvar exwm-workspace--window-y-offset)
-(defvar exwm-workspace--workareas)
 (declare-function exwm-layout--hide "exwm-layout.el" (id))
 (declare-function exwm-layout--iconic-state-p "exwm-layout.el" (&optional id))
 (declare-function exwm-layout--refresh "exwm-layout.el" ())
 (declare-function exwm-layout--show "exwm-layout.el" (id &optional window))
 (declare-function exwm-workspace--position "exwm-workspace.el" (frame))
 (declare-function exwm-workspace--update-offsets "exwm-workspace.el" ())
+(declare-function exwm-workspace--workarea "exwm-workspace.el" (frame))
 
-(defun exwm-floating--set-allowed-actions (id tilling)
-  "Set _NET_WM_ALLOWED_ACTIONS."
+(defun exwm-floating--set-allowed-actions (id tiled-p)
+  "Set _NET_WM_ALLOWED_ACTIONS for window with ID.
+If TILED-P is non-nil, set actions for tiled window."
   (exwm--log "#x%x" id)
   (xcb:+request exwm--connection
       (make-instance 'xcb:ewmh:set-_NET_WM_ALLOWED_ACTIONS
                      :window id
-                     :data (if tilling
+                     :data (if tiled-p
                                (vector xcb:Atom:_NET_WM_ACTION_MINIMIZE
                                        xcb:Atom:_NET_WM_ACTION_FULLSCREEN
                                        xcb:Atom:_NET_WM_ACTION_CHANGE_DESKTOP
@@ -169,6 +161,8 @@ This is also used by X window containers.")
                           (get-buffer "*scratch*")))
                   (make-frame
                    `((minibuffer . ,(minibuffer-window exwm--frame))
+                     (tab-bar-lines . 0)
+                     (tab-bar-lines-keep-state . t)
                      (left . ,(* window-min-width -10000))
                      (top . ,(* window-min-height -10000))
                      (width . ,window-min-width)
@@ -192,12 +186,8 @@ This is also used by X window containers.")
     (set-frame-parameter frame 'exwm-container frame-container)
     ;; Fix illegal parameters
     ;; FIXME: check normal hints restrictions
-    (let* ((workarea (elt exwm-workspace--workareas
-                          (exwm-workspace--position original-frame)))
-           (x* (aref workarea 0))
-           (y* (aref workarea 1))
-           (width* (aref workarea 2))
-           (height* (aref workarea 3)))
+    (with-slots ((x* x) (y* y) (width* width) (height* height))
+        (exwm-workspace--workarea original-frame)
       ;; Center floating windows
       (when (and (or (= x 0) (= x x*))
                  (or (= y 0) (= y y*)))
@@ -276,7 +266,8 @@ This is also used by X window containers.")
            (floating-mode-line (plist-get exwm--configurations
                                           'floating-mode-line))
            (floating-header-line (plist-get exwm--configurations
-                                            'floating-header-line)))
+                                            'floating-header-line))
+           (border-pixel (exwm--color->pixel exwm-floating-border-color)))
       (if floating-mode-line
           (setq exwm--mode-line-format (or exwm--mode-line-format
                                            mode-line-format)
@@ -323,15 +314,12 @@ This is also used by X window containers.")
                          :class xcb:WindowClass:InputOutput
                          :visual 0
                          :value-mask (logior xcb:CW:BackPixmap
-                                             (if exwm-floating--border-pixel
+                                             (if border-pixel
                                                  xcb:CW:BorderPixel 0)
-                                             xcb:CW:OverrideRedirect
-                                             (if exwm-floating--border-colormap
-                                                 xcb:CW:Colormap 0))
+                                             xcb:CW:OverrideRedirect)
                          :background-pixmap xcb:BackPixmap:ParentRelative
-                         :border-pixel exwm-floating--border-pixel
-                         :override-redirect 1
-                         :colormap exwm-floating--border-colormap))
+                         :border-pixel border-pixel
+                         :override-redirect 1))
       (xcb:+request exwm--connection
           (make-instance 'xcb:ewmh:set-_NET_WM_NAME
                          :window frame-container
@@ -479,7 +467,9 @@ This is also used by X window containers.")
     (select-frame-set-input-focus exwm-workspace--current)))
 
 (defun exwm-floating--start-moveresize (id &optional type)
-  "Start move/resize."
+  "Start move/resize for window with ID.
+When non-nil, TYPE indicates the type of move/resize.
+Float resizing is stopped when TYPE is nil."
   (exwm--log "#x%x" id)
   (let ((buffer-or-id (or (exwm--id->buffer id) id))
         frame container-or-id x y width height cursor)
@@ -685,7 +675,7 @@ This is also used by X window containers.")
     (setq exwm-floating--moveresize-calculate nil)))
 
 (defun exwm-floating--do-moveresize (data _synthetic)
-  "Perform move/resize."
+  "Perform move/resize on floating window with DATA."
   (when exwm-floating--moveresize-calculate
     (let* ((obj (make-instance 'xcb:MotionNotify))
            result value-mask x y width height buffer-or-id container-or-id)
@@ -758,39 +748,9 @@ Both DELTA-X and DELTA-Y default to 1.  This command should be bound locally."
                           nil nil))
     (xcb:flush exwm--connection)))
 
-(defun exwm-floating--init-border ()
-  "Initialize border colormap and pixel."
-  (exwm--log)
-  ;; Use the default colormap.
-  (unless exwm-floating--border-colormap
-    (with-slots (roots) (xcb:get-setup exwm--connection)
-      (with-slots (default-colormap) (car roots)
-        (setq exwm-floating--border-colormap default-colormap))))
-  ;; Free any previously allocated pixel.
-  (when exwm-floating--border-pixel
-    (xcb:+request exwm--connection
-        (make-instance 'xcb:FreeColors
-                       :cmap exwm-floating--border-colormap
-                       :plane-mask 0
-                       :pixels (vector exwm-floating--border-pixel)))
-    (setq exwm-floating--border-pixel nil))
-  ;; Allocate new pixel.
-  (let ((color (x-color-values (or exwm-floating-border-color "")))
-         reply)
-    (when color
-      (setq reply (xcb:+request-unchecked+reply exwm--connection
-                      (make-instance 'xcb:AllocColor
-                                     :cmap exwm-floating--border-colormap
-                                     :red (pop color)
-                                     :green (pop color)
-                                     :blue (pop color))))
-      (when reply
-        (setq exwm-floating--border-pixel (slot-value reply 'pixel))))))
-
 (defun exwm-floating--init ()
   "Initialize floating module."
   (exwm--log)
-  (exwm-floating--init-border)
   ;; Initialize cursors for moving/resizing a window
   (xcb:cursor:init exwm--connection)
   (setq exwm-floating--cursor-move
