@@ -24,9 +24,9 @@
 
 ;; This module adds system tray support for EXWM.
 
-;; To use this module, load and enable it as follows:
-;;   (require 'exwm-systemtray)
-;;   (exwm-systemtray-enable)
+;; To use this module, enable it as follows:
+;;
+;;   (exwm-systemtray-mode 1)
 
 ;;; Code:
 
@@ -50,6 +50,18 @@
   "System tray."
   :group 'exwm)
 
+;;;###autoload
+(define-minor-mode exwm-systemtray-mode
+  "Toggle EXWM systemtray support."
+  :global t
+  :group 'exwm-systemtray
+  (exwm--global-minor-mode-body systemtray))
+
+(defun exwm-systemtray-enable ()
+  "Enable EXWM systemtray support."
+  (exwm-systemtray-mode 1))
+(make-obsolete 'exwm-systemtray-enable "Use `exwm-systemtray-mode' instead." "0.30")
+
 (defcustom exwm-systemtray-height nil
   "System tray height.
 
@@ -87,7 +99,8 @@ TrueColor-24\" can be used to force Emacs to use 24-bit depth."
 using 32-bit depth.  Using `workspace-background' instead.")
            (setq value 'workspace-background))
          (set-default symbol value)
-         (when (and exwm-systemtray--connection
+         (when (and exwm-systemtray-mode
+                    exwm-systemtray--connection
                     exwm-systemtray--embedder-window)
            ;; Change the background color for embedder.
            (exwm-systemtray--set-background-color)
@@ -478,7 +491,9 @@ Argument DATA contains the raw event data."
 (cl-defun exwm-systemtray--init ()
   "Initialize system tray module."
   (exwm--log)
-  (cl-assert (not exwm-systemtray--connection))
+  ;; idempotent initialization
+  (when exwm-systemtray--connection
+    (cl-return-from exwm-systemtray--init))
   (cl-assert (not exwm-systemtray--list))
   (cl-assert (not exwm-systemtray--selection-owner-window))
   (cl-assert (not exwm-systemtray--embedder-window))
@@ -679,12 +694,6 @@ Argument DATA contains the raw event data."
     (when (boundp 'exwm-randr-refresh-hook)
       (remove-hook 'exwm-randr-refresh-hook #'exwm-systemtray--refresh-all))))
 
-(defun exwm-systemtray-enable ()
-  "Enable system tray support for EXWM."
-  (exwm--log)
-  (add-hook 'exwm-init-hook #'exwm-systemtray--init)
-  (add-hook 'exwm-exit-hook #'exwm-systemtray--exit))
-
 
 
 (provide 'exwm-systemtray)
